# Package placeholder

from .clock import Time, Clock, round_date
from .utils import dt2ts, ts2iso, iso2ts, tt2ts, dt2ts, tt2ts, ts2tt
from .timezone import UTC, TzLocal, TzTest, testTimeZone
from .strptime import strptime




# -*- coding: utf-8 -*-
"""
.. :doctest:

"""

import datetime
import time
import warnings
import dateutil.parser

from zope.interface import provider, implementer
from zope.component import queryUtility

from .interfaces import ITime, IClock
from .timezone import UTC, TzLocal
from .strptime import strptime
from .utils import dt2ts


try:
    unicode = unicode
except NameError:  ## pragma: no cover
    # 'unicode' is undefined, must be Python 3
    str = str
    unicode = str
    bytes = bytes
    basestring = (str,bytes)
else:  ## pragma: no cover
    # 'unicode' exists, must be Python 2
    str = str
    unicode = unicode
    bytes = str
    basestring = basestring


DEFAULT_PARSER_FORMATS = [
    "%Y-%m-%d",
    "%Y-%m-%d %H:%M",
    "%Y-%m-%d %H:%M:%S",
    "%m-%d",
    "%m/%d",
    "%m-%d %H:%M:%S",
    "%m-%d %H:%M",
    "%d %H:%M",
    "%d %Hh%M",
    "%d %Hh",
    "%H:%M:%S",
    "%H:%M",
    "%Hh%M",
    "%Hh",
    "%Mm",
    "%M",
    ]


def deprecation(message):
    warnings.warn(message, DeprecationWarning, stacklevel=2)


def deprecate(in_favor=None):

    def decorator(f):

        def wrapped(*args, **kwargs):
            if in_favor:
                deprecation("Call to %r is deprecated, use %r instead."
                            % (f.__name__, in_favor))
            else:
                deprecation("Call to %r is deprecated." % (f.__name__, ))
            return f(*args, **kwargs)
        return wrapped
    return decorator


def round_date(date):
    """Round a timedelta to the last minute (remove seconds and microseconds).

    Setup:

         >>> from sact.epoch import round_date, Time

    Nothing to do:

         >>> round_date(Time(2010, 1, 1, 1, 1, 0))
         <Time 2010-01-01 01:01:00+00:00>

    Round to 1 minute when we have 1 minute and 30 seconds:

         >>> round_date(Time(2010, 1, 1, 1, 1, 30))
         <Time 2010-01-01 01:01:00+00:00>

    """
    assert(isinstance(date, Time))

    if date.second or date.microsecond:
        return date - datetime.timedelta(seconds=date.second,
                                         microseconds=date.microsecond)
    return date


@implementer(IClock)
class Clock(object):
    """Time Factory

    Will only serve the current time.

    Usage
    =====

        >>> from sact.epoch.clock import Clock
        >>> c = Clock()

    We can use property 'ts' to get the timestamp and it should change very
    accurately at each call:

        >>> t1 = c.ts
        >>> t2 = c.ts
        >>> t1 < t2
        True

    If we need a full object we should use:

        >>> c.time
        <Time ...>

    """

    @property
    def time(self):
        """Should return later a Time object"""

        return Time.utcfromtimestamp(self.ts)

    @property
    def ts(self):
        return time.time()


@implementer(IClock)
class ManageableClock(Clock):
    r"""Creates a manageable time object

    Can be used to control what time it is. Start/Stop method can
    start/stop time, and wait/set will alter current time.

    Usage
    =====

        >>> from sact.epoch.clock import ManageableClock
        >>> mc = ManageableClock()

    Stopping time
    -------------

        >>> mc.stop()
        >>> t1 = mc.ts
        >>> t2 = mc.ts
        >>> assert t1 == t2, 'time %r should be equal to %r and it isn\'t' \
        ...              % (t1, t2)
        >>> mc.is_running
        False

    Stoping while running should do nothing:

        >>> mc.stop()
        >>> mc.is_running
        False


    Restarting time
    ---------------

        >>> mc.start()
        >>> t1 = mc.ts
        >>> t2 = mc.ts
        >>> assert t1 != t2, 'time %r should NOT be equal to %r and it is' \
        ...              % (t1, t2)
        >>> mc.is_running
        True

    Restarting while running should do nothing:

        >>> mc.start()
        >>> mc.is_running
        True

        >>> t3 = mc.ts
        >>> assert t1 < t3, \
        ...    'time %r should be superior to %r and it isn\'t' \
        ...    % (t3, t1)


    Setting time
    ------------

        >>> mc.stop()
        >>> mc.ts = 0
        >>> mc.ts
        0
        >>> mc.start()
        >>> ts = mc.ts
        >>> assert ts > 0, \
        ...    'clock should have been running and thus timestamp should be greater than 0.' \
        ...    'It was %r.' % ts

    Altering time
    -------------

        >>> mc.stop()
        >>> mc.ts = 20
        >>> mc.ts += 10
        >>> mc.ts
        30
        >>> mc.start()
        >>> ts = mc.ts
        >>> assert ts > 30, \
        ...    'clock should have been running and thus timestamp should be greater than 30.' \
        ...    'It was %r.' % ts

    Setting time should not stop the clock if it was running:

        >>> mc.ts = 20
        >>> mc.is_running
        True


    Altering with wait
    ------------------

        >>> mc.stop()
        >>> mc.ts = 0
        >>> mc.wait(minutes=5)
        >>> mc.ts
        300
        >>> mc.start()
        >>> mc.wait(minutes=5)
        >>> ts = mc.ts
        >>> assert ts > 600, \
        ...    'clock should have been running and thus timestamp should be greater than 600.' \
        ...    'It was %r.' % ts

    With no specific argument, it'll be casted to int and interpreted
    in seconds:

        >>> mc.stop()
        >>> mc.ts = 0
        >>> mc.wait(3)
        >>> mc.ts
        3


    """

    def __init__(self):
        self.delta = 0
        self._ft = None ## freezed time

    def start(self):
        if self.is_running:
            return
        ## use _ft to calculate the time delta
        self.delta = self.ts - self.delta - self._ft
        self._ft = None

    def stop(self):
        ## save real current time
        self._ft = self.ts - self.delta

    @property
    def is_running(self):
        return self._ft is None

    def get_ts(self):
        if not self.is_running:
            return self._ft
        return time.time() + self.delta

    def set_ts(self, value):
        self.delta = value - time.time()
        # don't forget to update self._ft
        if self._ft is not None:
            self._ft = value

    ts = property(get_ts, set_ts)

    def wait(self, timedelta=None, **kwargs):
        """Provide a convenient shortcut to alter the current time

        timedelta can be an int/float or a timedelta objet from timedelta

        """
        if timedelta is None:
            timedelta = datetime.timedelta(**kwargs)

        if isinstance(timedelta, datetime.timedelta):
            secs = timedelta.days * 86400 + timedelta.seconds
        else:
            secs = int(timedelta)

        self.ts += secs


DefaultClock = Clock()
DefaultManageableClock = ManageableClock()

@provider(ITime)
class Time(datetime.datetime):
    """Time Factory

      Time is an abstraction of a specific point of time. As a subclass of
    datetime, it provides the same functionality. Additionaly it ensures that a
    timezone is always associated to the Time instance, and some simple common
    representation are provided as properties.

      And most important: Time.now() will silently request for the registered
    clock to get time information, allowing to manage time easily.

      Same mecanism is used to get the local time zone, allowing to override
    the detection of the local time zone when using Time.now_lt().


    Limitations
    ===========

      Underlying Datetime object have some methods that can fail after
    initialisation as Datetime.strftime(). Time object make an extensive
    use of some of them and thus will perform a validity check of these
    methods at initialisation time. This will limit the possible range
    of times that is supported currently by the Time object.

      The earlier date possible depends of your architecture. You can set the
    'disable_validity_check' optional argument to 'True' at init time to bypass
    the checkings but notice that some localized methods are then not garanteed
    to work and have reportedly failed on dates earlier than year 1900::

    On python 2:

      # >>> Time(1, 1, 1)
      # Traceback (most recent call last):
      # ...
      # ValueError: Encountered datetime method limitation: ...

    But this does not occur on python 3:

      # >>> Time(1, 1, 1)
      # <Time 0001-01-01 00:00:00+00:00>

    Usage
    =====

    This is quite straightforward to use:

        >>> from sact.epoch import Time
        >>> Time.now()
        <Time ...+00:00>

    Notice that it has a timezone information set. Silently, the current time
    was asked to the registered clock available, which is by default the normal
    clock.


    We can give a better view thanks to a manageable clock as time reference:

        >>> from sact.epoch.clock import ManageableClock
        >>> clock = ManageableClock()

    We will stop the time to epoch:

        >>> clock.stop()
        >>> clock.ts = 0

    Let's set it as reference:

        >>> from zope.component import globalSiteManager as gsm
        >>> gsm.registerUtility(clock)

    Now, let's set our TzTest as local timezone, remember it has 5 minutes
    difference to UTC:

        >>> from sact.epoch import testTimeZone
        >>> from sact.epoch.interfaces import ITimeZone

        >>> gsm.registerUtility(testTimeZone, ITimeZone, name='local')

    Here is the result of each function:

        >>> Time.now()
        <Time 1970-01-01 00:00:00+00:00>

        >>> Time.now_lt()
        <Time 1970-01-01 00:05:00+00:05>

    Please note that there are 5 minutes of diff to UTC


    Instanciation
    =============

    It takes same arguments than datetime legacy object:

        >>> Time(1980, 1, 1)
        <Time 1980-01-01 00:00:00+00:00>

    Notice that in this case, it takes the UTC timezone.

    Additionnaly it can take a real datetime as argument:

        >>> from datetime import datetime
        >>> d = datetime(1970, 1, 1, tzinfo=testTimeZone)
        >>> t = Time(d)
        >>> t
        <Time 1970-01-01 00:00:00+00:05>

        >>> Time(t)
        <Time 1970-01-01 00:00:00+00:05>

    Or even a direct string:

        >>> Time('1970-01-01 00:00:00+00:05')
        <Time 1970-01-01 00:00:00+00:05>

        >>> Time(u'2000-01-01 00:00:00+01:00')
        <Time 2000-01-01 00:00:00+01:00>

    However, if you don't provide the timezone in the string representation
    or in the datetime (which is then called naive-datetimes):

        >>> Time(datetime(1970, 1, 1))
        Traceback (most recent call last):
        ...
        ValueError: No timezone hinted, nor found.

        >>> Time('1970-01-01 00:00:00')
        Traceback (most recent call last):
        ...
        ValueError: No timezone hinted, nor found.

    Hopefully, you can provide an hint to what is the intended timezone:

        >>> from sact.epoch import UTC
        >>> Time(datetime(1970, 1, 1), hint_src_tz=UTC())
        <Time 1970-01-01 00:00:00+00:00>
        >>> Time('2000-01-01 00:00:00', hint_src_tz=UTC())
        <Time 2000-01-01 00:00:00+00:00>

    Additionaly, partial date string parsing is available:

        >>> Time('15:30', hint_src_tz=UTC())
        <Time 1970-01-01 15:30:00+00:00>

    Avoid using more than one argument when providing a string, and be
    careful for instance to name the keyword argument::

        >>> Time('15:30', UTC())
        Traceback (most recent call last):
        ...
        SyntaxError: Too much positional arguments when using Time instanciation by string.


    Representations
    ^^^^^^^^^^^^^^^

    There are several standard representation that are available:

        >>> t.iso
        '1970-01-01 00:00:00+00:05'

    Short local (remove time zone):

        >>> t.short
        '1970-01-01 00:00:00'

    Short short local (remove seconds):

        >>> t.short_short
        '1970-01-01 00:00'

    timetuples:

        >>> t.tt
        time.struct_time(tm_year=1970, tm_mon=1, tm_mday=1, tm_hour=0, tm_min=0, tm_sec=0, tm_wday=3, tm_yday=1, tm_isdst=0)

    These are DEPRECATED usage::

        >>> import warnings
        >>> warnings.simplefilter('always')
        >>> t.iso_local
        '1970-01-01 00:00:00+00:05'

    Short local (remove time zone):

        >>> t.short_local
        '1970-01-01 00:00:00'

    Short short local (remove seconds):

        >>> t.short_short_local
        '1970-01-01 00:00'


    Timezone changing
    ^^^^^^^^^^^^^^^^^

    You can change the timezone (without changing the actual stored
    time) of a Time instance by using ``astimezone()`` which stems
    out underlying datetime API. We provide here 2 handy shortcuts:

        >>> Time(1980, 1, 1).local
        <Time 1980-01-01 00:05:00+00:05>
        >>> Time(1980, 1, 1).utc
        <Time 1980-01-01 00:00:00+00:00>

    These are DEPRECATED usage::

        >>> Time(1980, 1, 1).aslocal
        <Time 1980-01-01 00:05:00+00:05>
        >>> Time(1980, 1, 1).asutc
        <Time 1980-01-01 00:00:00+00:00>


    """

    def __new__(cls, *args, **kwargs):
        if len(args) and isinstance(args[0], datetime.datetime):
            return Time.from_datetime(args[0], **kwargs)

        if len(args) and isinstance(args[0], basestring):
            if len(args) > 1:
                raise SyntaxError(
                    "Too much positional arguments when using Time "
                    "instanciation by string.")
            if "hint_src_tz" in kwargs:
                try:
                    return Time.from_string(args[0], **kwargs)
                except ValueError:
                    pass

            if "relative" in kwargs:
                default = kwargs["relative"]
                del kwargs["relative"]
            else:
                ## XXXvlab: hum, we have to craft a Time without timezone
                ## to force error if no explicit timezone is given to the
                ## Time instanciation
                default = Time.now().replace(tzinfo=kwargs.get("hint_src_tz"))

            return Time(dateutil.parser.parse(args[0], default=default),
                        **kwargs)

        if 'tzinfo' not in kwargs and len(args) < 8:
            # XXXjballet: to test
            kwargs['tzinfo'] = UTC()

        check = not kwargs.pop('disable_validity_check', False)

        dt = super(Time, cls).__new__(cls, *args, **kwargs)

        if check:
            ## Check that the architecture can call localized methods
            try:
                dt.strftime('%Y')
            except ValueError as e:  ## pragma: no cover
                raise ValueError("Encountered datetime method limitation:"
                                 " %s" % str(e))

        return dt

    def __repr__(self):
        return "<Time %s>" % self

    @classmethod
    def from_datetime(cls, dt, hint_src_tz=None):
        """Convert a datetime object with timezone to a Time object

        This method provides a handy way to convert datetime objects to Time
        objects:

            >>> import datetime
            >>> from sact.epoch import UTC, TzLocal
            >>> dt = datetime.datetime(2000, 1, 1, tzinfo=UTC())
            >>> Time.from_datetime(dt)
            <Time 2000-01-01 00:00:00+00:00>

        The provided datetime should contain a timezone information or the
        conversion will fail:

            >>> Time.from_datetime(datetime.datetime.now())
            Traceback (most recent call last):
            ...
            ValueError: No timezone hinted, nor found.

        You can provide yourself this timezone information thanks to
        the 'hint_src_tz' argument:

            >>> Time.from_datetime(datetime.datetime(2000, 1, 1),
            ...     hint_src_tz=UTC())
            <Time 2000-01-01 00:00:00+00:00>

        However, this remains an hint, and if timezone can be found in
        the datetime, the 'hint_src_tz' argument will be ignored:

            >>> Time.from_datetime(
            ...      datetime.datetime(2000, 1, 1, tzinfo=UTC()),
            ...      hint_src_tz=TzLocal())
            <Time 2000-01-01 00:00:00+00:00>


        """

        tzinfo = dt.tzinfo if dt.tzinfo else hint_src_tz
        if tzinfo is None:
            raise ValueError("No timezone hinted, nor found.")

        return cls(dt.year, dt.month, dt.day, dt.hour,
                   dt.minute, dt.second, dt.microsecond,
                   tzinfo)

    def __add__(self, delta):
        """Override datetime '+' to return a Time object

        Add a timedelta:

            >>> Time(2010, 1, 1) + datetime.timedelta(days=1)
            <Time 2010-01-02 00:00:00+00:00>

        Add an other Time (send the original exception):

            >>> Time(2010, 1, 1) + Time(1970, 1, 1)
            Traceback (most recent call last):
            ...
            TypeError: unsupported operand type(s) for +: 'Time' and 'Time'

        """
        d = super(Time, self).__add__(delta)
        if isinstance(d, datetime.datetime):
            return self.from_datetime(d)
        return d

    def __sub__(self, delta):
        """Override datetime '-' to return a Time object

        Sub a timedelta:

            >>> Time(2010, 1, 1) - datetime.timedelta(days=1)
            <Time 2009-12-31 00:00:00+00:00>

        Sub an other Time:

            >>> Time(2010, 1, 2) - Time(2010, 1, 1)
            datetime.timedelta(1)

        """
        d = super(Time, self).__sub__(delta)
        if isinstance(d, datetime.datetime):
            return self.from_datetime(d)
        return d

    @staticmethod
    def now():
        utility = queryUtility(IClock, default=DefaultClock)
        return utility.time.replace(tzinfo=UTC())

    ## XXXvlab: to deprecate
    @staticmethod
    def now_lt():
        return Time.now().local

    @classmethod
    def utcfromtimestamp(cls, ts):
        """Return a UTC datetime from a timestamp.

            >>> Time.utcfromtimestamp(0)
            <Time 1970-01-01 00:00:00+00:00>

        """

        dt = super(Time, cls).utcfromtimestamp(ts)
        return dt.replace(tzinfo=UTC())

    @classmethod
    def fromtimestamp(cls, ts, tz=None):
        """Return a UTC datetime from a timestamp.

            >>> from sact.epoch.timezone import testTimeZone
            >>> Time.fromtimestamp(0, tz=testTimeZone)
            <Time 1970-01-01 00:05:00+00:05>

        """
        t = cls.utcfromtimestamp(ts)
        if tz is not None:
            t = t.astimezone(tz)
        return t

    @classmethod
    def from_string(cls, date_str, hint_src_tz,
                    relative=True, formats=None):
        """Return a UTC datetime from a possibily partial string representation

        By default, output is relative to localtime:

            >>> Time.from_string('15h30', hint_src_tz=UTC())
            <Time 1970-01-01 15:30:00+00:00>

        But it can be specified of course:

            >>> t = Time.from_string('2000-01-01 00:00:00', hint_src_tz=UTC())
            >>> Time.from_string('15h30', hint_src_tz=UTC(),
            ...     relative=t)
            <Time 2000-01-01 15:30:00+00:00>

        """
        formats = formats or DEFAULT_PARSER_FORMATS
        for f in formats:
            try:
                return cls.strptime(
                    date_str, f, hint_src_tz=hint_src_tz,
                    relative=relative)
            except ValueError:
                pass
        raise ValueError("No format seems to know how to parse your string %r"
                         % (date_str, ))

    @classmethod
    def strptime(cls, value, format, hint_src_tz, relative=False):
        """Parse a string to create a Time object.

        hint_src_tz is the source time zone in which the value
        should be interpreted.

        The result Time instance will be in UTC (but you can move it
        with astimezone, or the shortcut '.local').

        Usage
        =====

        source timezone
        ^^^^^^^^^^^^^^^

            >>> from sact.epoch import UTC, TzTest
            >>> Time.strptime('2000-01-01', '%Y-%m-%d', hint_src_tz=UTC())
            <Time 2000-01-01 00:00:00+00:00>

        If source is UTC, then it's the same as the output. No surprise here.

            >>> from sact.epoch.timezone import testTimeZone as ttz

            >>> t0 = Time.strptime('2000-01-01', '%Y-%m-%d',
            ...                    hint_src_tz=ttz)
            >>> t0
            <Time 1999-12-31 23:55:00+00:00>

        If source is in testTimeZone time, then UTC is 5 minutes
        before. The output is by default in UTC. So it's prettier if
        we look at this time in it's original timezone::

            >>> t0.astimezone(ttz)
            <Time 2000-01-01 00:00:00+00:05>

        relative
        ^^^^^^^^

        There's a ``relative`` option which if non-False, will fill
        non given time with the given time instance as relative (or now in
        local time zone if the ``relative`` value is set to True)::

            >>> clock = queryUtility(IClock)

            >>> t1 = Time.strptime('2000-01-01', '%Y-%m-%d', UTC())
            >>> clock.ts = t1.timestamp

        We've just jumped in time and are now at ``t1``. So if we now ask for
        ``strptime()`` in relative mode with only hours and minutes given::

            >>> Time.strptime('15:08', '%H:%M', UTC(), relative=True)
            <Time 2000-01-01 15:08:00+00:00>
            >>> Time.strptime('15:08', '%H:%M', ttz, relative=True)
            <Time 2000-01-01 15:03:00+00:00>

        We could have also set the relative time directly as a value in the
        ``relative`` argument::

            >>> t2 = Time.strptime('1990-05-05', '%Y-%m-%d', UTC())
            >>> Time.strptime('15:08', '%H:%M', ttz, relative=t2)
            <Time 1990-05-05 15:03:00+00:00>

        You can notice how that is different from default behavior, which
        takes EPOCH as relative::

            >>> Time.strptime('15:08', '%H:%M', UTC(), relative=False)
            <Time 1900-01-01 15:08:00+00:00>

        """
        if relative is False:
            dt = super(Time, cls).strptime(value, format)
            dt = dt.replace(tzinfo=hint_src_tz)
            return dt.utc

        relative = Time.now() if relative is True else relative

        input_time = relative.timetuple(), relative.microsecond
        time_struct, microseconds = strptime(value, format,
                                             reference=input_time)
        dt = Time(*time_struct[:6])
        dt = dt.replace(microsecond=microseconds, tzinfo=hint_src_tz)
        return dt.utc

    def astimezone(self, tz):
        """Convert Time object to another timezone and return a Time object

        This overrides the datetime's method to return a Time object instead of
        a datetime object:

            >>> from sact.epoch.timezone import testTimeZone
            >>> type(Time.now().astimezone(testTimeZone))
            <class 'sact.epoch...Time'>

        """

        dt = super(Time, self).astimezone(tz)
        return self.from_datetime(dt)

    @property
    def timestamp(self):
        """Convert this Time instance in a unix timestamp in UTC

        See sact.epoch.utils

            >>> Time(1970, 1, 1, 0, 0, 1).timestamp
            1

        """
        return dt2ts(self)

    @property
    def utc(self):
        return self.astimezone(UTC())

    @property
    def local(self):
        return self.astimezone(TzLocal())

    @property
    def iso(self):
        """Return the iso format in local time

            >>> Time(1970, 1, 1, 1, 1).iso
            '1970-01-01 01:01:00+00:00'

        """
        return self.isoformat(" ")

    @property
    def short(self):
        """Return as iso with without time zone

        Warning, this looses the time zone information and thus could
        be dangerous. Make sure that if you store this, the time zone is
        clearly inferable from its environment.

            >>> Time(1970, 1, 1, 1, 1).short
            '1970-01-01 01:01:00'

        """
        return self.strftime('%Y-%m-%d %H:%M:%S')

    @property
    def short_short(self):
        """Idem without seconds

        Warning, this looses the time zone information and thus could
        be dangerous. Make sure that if you store this, the time zone is
        clearly inferable from its environment.

            >>> Time(1970, 1, 1, 1, 1).short_short
            '1970-01-01 01:01'

        """
        return self.strftime('%Y-%m-%d %H:%M')

    @property
    def tt(self):
        """Returns the timetuple of current Time instance"""
        return self.timetuple()

    ts = timestamp

    ##
    ## DEPRECATED
    ##

    @property
    @deprecate(in_favor='local')
    def aslocal(self):
        return self.local

    @property
    @deprecate(in_favor='utc')
    def asutc(self):
        return self.utc

    @deprecate(in_favor='.local.strftime')
    def strftime_local(self, *args, **kwargs):
        return self.local.strftime(*args, **kwargs)

    @property
    @deprecate(in_favor='.local.iso')
    def iso_local(self):
        return self.local.isoformat(" ")

    @property
    @deprecate(in_favor='.local.short')
    def short_local(self):
        return self.local.strftime('%Y-%m-%d %H:%M:%S')

    @property
    @deprecate(in_favor='.local.short_short')
    def short_short_local(self):
        return self.local.strftime('%Y-%m-%d %H:%M')


"""
Let's unregister the test Timezone and test Clock:

    >>> gsm.unregisterUtility(clock)
    True

    >>> gsm.unregisterUtility(testTimeZone, ITimeZone, 'local')
    True

"""




from zope.interface import Interface, Attribute


class ITimeZone(Interface):
    """Standard TimeZone interface as defined in datetime module"""

    def utcoffset(dt):
        """Return offset of local time from UTC, in minutes"""

    def dst(dt):
        """Return the daylight saving time (DST) adjustment, in minutes"""

    def tzname(dt):
        """Return the time zone name corresponding to the datetime object dt"""


class ITime(Interface):
    """Factory to make time object.

    minimal interface

    """

    def now():
        """Return a time object that represent the current time"""


class IClock(Interface):
    """Factory to make time object.

    minimal interface

    """

    def time():
        """Return a time object that represent the current time"""


class IManageableClock(IClock):

    is_running = Attribute(
        u"Freeze the return result of now() command")

    def set(date):
        """Set the result of now() command to date"""

    def stop():
        """Freeze the return result of now() command"""

    def start():
        """Unfreeze the return result of now() command"""

    def wait(timelapse=None, days=0, hours=0, minutes=0, seconds=0):
        """Shortcut to alter_now relative

        Should accept negative value also.

        """



import datetime
import time

from .interfaces import ITimeZone

from zope.interface import implementer
from zope.component import queryUtility


ZERO = datetime.timedelta(seconds=0)


def is_dst(dt):
    """Return True or False depending of tm_isdst value

    Convert dt in timestamp, and get time object with this timestamp to get
    the localtime and see if this time is dst or not

    """

    tt = (dt.year, dt.month, dt.day,
          dt.hour, dt.minute, dt.second,
          dt.weekday(), 0, -1)
    stamp = time.mktime(tt)

    tt = time.localtime(stamp)
    return tt.tm_isdst > 0


@implementer(ITimeZone)
class UTC(datetime.tzinfo):
    """Represent the UTC timezone.

    From http://docs.python.org/library/datetime.html#tzinfo-objects examples

    XXXvlab: pytz.utc isn't better ?

    """

    def utcoffset(self, dt):
        return ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return ZERO

    def __repr__(self):
        return "<TimeZone: UTC>"


@implementer(ITimeZone)
class TzSystem(datetime.tzinfo):
    """Get the timezone locale of the system. It is used for datetime object.

    This object get the local DST and utcoffset.

    More explanation about how it is work for utcoffset:

    time.daylight is Nonzero if a DST timezone is defined.

    In this case we have two different values in stdoffset and dstoffset.

    For example for timezone 'Europe/Paris' we have:

    stdoffset = -3600
    dstoffset = -7200

    In `utcoffset` method, we check for the daylight saving or not and adjust
    offset in consequence.

    """

    # Get the right offset with DST or not
    stdoffset = datetime.timedelta(seconds=(- time.timezone))
    if time.daylight:
        dstoffset = datetime.timedelta(seconds=(- time.altzone))
    else:
        dstoffset = stdoffset

    # Get the DST adjustement in minutes
    dstdiff = dstoffset - stdoffset

    def utcoffset(self, dt):
        """Return offset of local time from UTC, in minutes"""

        return self.dstoffset if is_dst(dt) else self.stdoffset

    def dst(self, dt):
        """Return the daylight saving time (DST) adjustment, in minutes"""

        return self.dstdiff if is_dst(dt) else ZERO

    def tzname(self, dt):
        """Return time zone name of the datetime object dt"""

        return time.tzname[is_dst(dt)]

    def __repr__(self):
        return "<TimeZone: System>"


@implementer(ITimeZone)
class TzTest(datetime.tzinfo):
    """Timezone crafted for tests"""

    def utcoffset(self, dt):
        return datetime.timedelta(hours=0, minutes=5)

    def tzname(self, dt):
        return "GMT +5m"

    def dst(self, dt):
        return ZERO

    def __repr__(self):
        return "<TimeZone: Test>"


testTimeZone = TzTest()
defaultLocalTimeZone = TzSystem()


def TzLocal():
    """Get local timezone with ZCA"""

    return queryUtility(ITimeZone, name='local', default=defaultLocalTimeZone)



from setuptools import setup, find_packages

setup(
    name='sact.epoch',
    version="1.3.0",
    long_description=open("docs/source/overview.rst").read() + open("docs/source/changelog.rst").read(),
    description="Time object subclassing datetime allowing diverting local clock mecanism",
    # Get more strings from http://www.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Topic :: Software Development",
    ],
    keywords='sact time timedelta',
    author='SecurActive SA',
    author_email='opensource@securactive.net',
    url='http://github.com/securactive/sact.epoch',
    license='BSD License',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    namespace_packages=['sact'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        # -*- Extra requirements: -*-
        'python-dateutil',
        'zope.interface',
        'zope.component',
        'pytz',
    ],
    extras_require={'test': [
        'zope.testing',
        'zope.testrunner',
        # -*- Extra requirements: -*-
        'z3c.testsetup',
    ],
    },
    entry_points="""
    # -*- Entry points: -*-
    """,
)



# -*- coding: utf-8 -*-
"""
.. :doctest:

"""

import time
import _strptime
from datetime import date as datetime_date


def first(l, predicate):
    for e in l:
        if predicate(e):
            return e
    raise ValueError("No element in list matches given predicate.")


## Copy of ``_strptime()`` from python 2.7.3 (default, Apr 20 2012, 22:39:59)
## except for optional argument ``reference`` so as to add the option of a
## default ground time.

def strptime(data_string, format="%a %b %d %H:%M:%S %Y",
             reference=(1900, 1, 1, 0, 0, 0, -1, -1, -1),
             complete_with_zeroes=True):
    """Return time struct and microseconds based on an input and format string

    An optional ``reference`` is set by default to 1900-01-01 00:00:00.

    ``reference`` field is a tuple with:

    (year, month, day, hour, minute, second, weekday, julian, tz), microseconds

    Please note that weekday and julian will be ignored.

        >>> from sact.epoch.strptime import strptime
        >>> strptime('13:05', '%H:%M',
        ...          reference=((2000, 1, 1, 0, 0, 30, -1, -1, -1), 5))
        (time.struct_time(tm_year=2000, tm_mon=1, tm_mday=1, tm_hour=13, tm_min=5, tm_sec=0, tm_wday=5, tm_yday=1, tm_isdst=-1), 0)

    Notice how all the left most values (the biggest weight value)
    where used from the given reference, and how the left most value
    (least weight values) was zeroed. The middle part is what was
    parsed.

    You can switch of the zeroing of the left most part:

        >>> strptime('13:05', '%H:%M',
        ...          reference=((2000, 1, 1, 0, 0, 0, -1, -1, -1), 5),
        ...          complete_with_zeroes=False)
        (time.struct_time(tm_year=2000, tm_mon=1, tm_mday=1, tm_hour=13, tm_min=5, tm_sec=0, tm_wday=5, tm_yday=1, tm_isdst=-1), 5)


    """
    with _strptime._cache_lock:
        if _strptime._getlang() != _strptime._TimeRE_cache.locale_time.lang:
            _strptime._TimeRE_cache = TimeRE()
            _strptime._regex_cache.clear()
        if len(_strptime._regex_cache) > _strptime._CACHE_MAX_SIZE:
            _strptime._regex_cache.clear()
        locale_time = _strptime._TimeRE_cache.locale_time
        format_regex = _strptime._regex_cache.get(format)
        if not format_regex:
            try:
                format_regex = _strptime._TimeRE_cache.compile(format)
            # KeyError raised when a bad format is found; can be specified as
            # \\, in which case it was a stray % but with a space after it
            except KeyError as err:
                bad_directive = err.args[0]
                if bad_directive == "\\":
                    bad_directive = "%"
                del err
                raise ValueError("'%s' is a bad directive in format '%s'" %
                                    (bad_directive, format))
            # IndexError only occurs when the format string is "%"
            except IndexError:
                raise ValueError("stray %% in format '%s'" % format)
            _strptime._regex_cache[format] = format_regex
    found = format_regex.match(data_string)
    if not found:
        raise ValueError("time data %r does not match format %r" %
                         (data_string, format))
    if len(data_string) != found.end():
        raise ValueError("unconverted data remains: %s" %
                          data_string[found.end():])
    (year, month, day, hour, minute, second, weekday, julian, tz), \
           fraction = reference
    # Force calculation for julian and weekday
    julian = -1
    weekday = -1
    # Default to -1 to signify that values not known; not critical to have,
    # though
    week_of_year = -1
    week_of_year_start = -1
    found_dict = found.groupdict()
    for group_key in found_dict.keys():
        # Directives not explicitly handled below:
        #   c, x, X
        #      handled by making out of other directives
        #   U, W
        #      worthless without day of the week
        if group_key == 'y':
            year = int(found_dict['y'])
            # Open Group specification for strptime() states that a %y
            #value in the range of [00, 68] is in the century 2000, while
            #[69,99] is in the century 1900
            if year <= 68:
                year += 2000
            else:
                year += 1900
        elif group_key == 'Y':
            year = int(found_dict['Y'])
        elif group_key == 'm':
            month = int(found_dict['m'])
        elif group_key == 'B':
            month = locale_time.f_month.index(found_dict['B'].lower())
        elif group_key == 'b':
            month = locale_time.a_month.index(found_dict['b'].lower())
        elif group_key == 'd':
            day = int(found_dict['d'])
        elif group_key == 'H':
            hour = int(found_dict['H'])
        elif group_key == 'I':
            hour = int(found_dict['I'])
            ampm = found_dict.get('p', '').lower()
            # If there was no AM/PM indicator, we'll treat this like AM
            if ampm in ('', locale_time.am_pm[0]):
                # We're in AM so the hour is correct unless we're
                # looking at 12 midnight.
                # 12 midnight == 12 AM == hour 0
                if hour == 12:
                    hour = 0
            elif ampm == locale_time.am_pm[1]:
                # We're in PM so we need to add 12 to the hour unless
                # we're looking at 12 noon.
                # 12 noon == 12 PM == hour 12
                if hour != 12:
                    hour += 12
        elif group_key == 'M':
            minute = int(found_dict['M'])
        elif group_key == 'S':
            second = int(found_dict['S'])
        elif group_key == 'f':
            s = found_dict['f']
            # Pad to always return microseconds.
            s += "0" * (6 - len(s))
            fraction = int(s)
        elif group_key == 'A':
            weekday = locale_time.f_weekday.index(found_dict['A'].lower())
        elif group_key == 'a':
            weekday = locale_time.a_weekday.index(found_dict['a'].lower())
        elif group_key == 'w':
            weekday = int(found_dict['w'])
            if weekday == 0:
                weekday = 6
            else:
                weekday -= 1
        elif group_key == 'j':
            julian = int(found_dict['j'])
        elif group_key in ('U', 'W'):
            week_of_year = int(found_dict[group_key])
            if group_key == 'U':
                # U starts week on Sunday.
                week_of_year_start = 6
            else:
                # W starts week on Monday.
                week_of_year_start = 0
        elif group_key == 'Z':
            # Since -1 is default value only need to worry about setting tz if
            # it can be something other than -1.
            found_zone = found_dict['Z'].lower()
            for value, tz_values in enumerate(locale_time.timezone):
                if found_zone in tz_values:
                    # Deal with bad locale setup where timezone names are the
                    # same and yet time.daylight is true; too ambiguous to
                    # be able to tell what timezone has daylight savings
                    if (time.tzname[0] == time.tzname[1] and
                       time.daylight and found_zone not in ("utc", "gmt")):
                        break
                    else:
                        tz = value
                        break
    # If we know the week of the year and what day of that week, we can figure
    # out the Julian day of the year.
    if julian == -1 and week_of_year != -1 and weekday != -1:
        week_starts_Mon = True if week_of_year_start == 0 else False
        julian = _strptime._calc_julian_from_U_or_W(year, week_of_year, weekday,
                                            week_starts_Mon)
    if complete_with_zeroes:
        idxs = ("j", "Y", "mBb", "UW", "dAaw", "HI", "M", "S", "f")
        specified = [idxs.index(gk)
                     for gk in [first(idxs, lambda key_group: k in key_group)
                                for k in found_dict.keys()]]
        rightmost_specified = max(specified) if len(specified) else 0
        base = (1900, 1, 1, 1, 0, 0, 0, 0)
        values = year, month, week_of_year, day, hour, minute, second, fraction
        year, month, week_of_year, day, hour, minute, second, fraction = \
               values[:rightmost_specified] + base[rightmost_specified:]

    # Cannot pre-calculate datetime_date() since can change in Julian
    # calculation and thus could have different value for the day of the week
    # calculation.
    if julian == -1:
        # Need to add 1 to result since first day of the year is 1, not 0.
        julian = datetime_date(year, month, day).toordinal() - \
                  datetime_date(year, 1, 1).toordinal() + 1
    else:  # Assume that if they bothered to include Julian day it will
           # be accurate.
        datetime_result = datetime_date.fromordinal(
            (julian - 1) + datetime_date(year, 1, 1).toordinal())
        year = datetime_result.year
        month = datetime_result.month
        day = datetime_result.day
    if weekday == -1:
        weekday = datetime_date(year, month, day).weekday()
    return (time.struct_time((year, month, day,
                              hour, minute, second,
                              weekday, julian, tz)), fraction)



# -*- coding: utf-8 -*-
"""
.. :doctest:

These small functions deal with 3 time representation without timezone
information.  It is probably simpler to think them as being all UTC.

ISO (abbreviated 'iso')

    Which is a string representation. ie: '2008-11-01 10:00:00'

time tuples (abbreviated 'tt')

    Which are 9-tuples of the form (ie: (2008, 11, 01, 10, 0, 0))

timestamp (abbreviated 'ts')

    Which is the number of seconds since EPOCH. (ie: 1225533600)


"""

import time
import datetime
import calendar


def dt2ts(dt):
    """Converts a datetime object to timestamp

    Important note: naive datetime are supported and are considered
    UTC.

    Usage
    =====

        >>> from sact.epoch import dt2ts
        >>> from sact.epoch import tt2ts, TzLocal, UTC
        >>> import datetime

        >>> epoch = datetime.datetime(1970,1,1)
        >>> dt2ts(epoch)
        0

    This works also with non-naive datetime, and will output
    the right timestamp.

        >>> dt = epoch.replace(tzinfo=UTC())
        >>> dt2ts(dt)
        0

    ``.astimezone()`` will change only the timezone, keeping the real
    moment in time. So timestamp should not change:

        >>> dt2ts(dt.astimezone(TzLocal()))
        0

    """
    return int(tt2ts(dt.utctimetuple()))


def ts2iso(ts):
    """Returns an (UTC) ISO representation of a timestamp

        >>> from sact.epoch import iso2ts, ts2iso

        >>> ts2iso(iso2ts('2008-11-01 10:00:00'))
        '2008-11-01 10:00:00'

    """
    return datetime.datetime.utcfromtimestamp(ts).isoformat(" ")


def iso2ts(iso):
    """Returns a timestamp from an (UTC) ISO representation

    Note that it doesn't take into account local representation.

        >>> from sact.epoch import iso2ts, ts2tt

    EPOCH should return 0 as expected:

        >>> iso2ts('1970-01-01 00:00:00')
        0

    And with any date, we should be able to reconstruct it:

        >>> ts2tt(iso2ts('2008-12-06 12:17:31'))[0:6]
        (2008, 12, 6, 12, 17, 31)

    """

    return tt2ts(time.strptime(iso + " UTC", "%Y-%m-%d %H:%M:%S %Z"))


tt2ts = calendar.timegm
ts2tt = time.gmtime
